/*
   Parser.java -- database lookup

   $Id: Parser.java,v 1.1 1999/05/24 19:41:57 eg Exp $

   Copyright (c) 1995-1999 The ht://Dig Group
   Copyright (C) 1999, Edwin Glaser

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

package htsearch;

import com.sleepycat.db.*;
import java.util.Vector;
import java.util.Enumeration;

class Stack extends Vector
{
  Object pop()
  {
	int last = size() - 1;
	if ( last < 0 )
	{
		return null;
	}
	Object tmp = elementAt(last);
	removeElementAt(last);
	return tmp;
  }

  void push(Object obj)
  {
	addElement(obj);
  }
}

class ResultList extends Vector
{
  boolean isIgnore;

  DocMatch find (int id)
  {
	for ( int i = 0; i < size(); i++ )
	{
		DocMatch m = (DocMatch) elementAt(i);
		if ( m.id == id )
		{
			return m;
		}
	}
	return null;
  }
}

/**
 * Parser -- database lookup
 *
 *
 */

public class Parser extends DbEnv
{

  static final int WORD = 1000;
  static final int DONE = 1001;

  static final int MAX_WORD_LENGTH = 12;

  Vector		tokenVector;
  Vector		searchWords;
  Enumeration   tokens;
  WeightWord	current;
  int			lookahead;
  Stack		    stack;
  Db	        wordDb;
  Db	        indexDb;
  Db	        docDb;

  boolean	    valid;
  String		error;

  public Parser(String wordFile, String indexFile, String docFile)
	throws DbException
  {
	super(null, null, 0);
	error = "Can't open database";

	wordDb = Db.open(wordFile, Db.DB_BTREE, Db.DB_RDONLY, 0, this, null);
	indexDb = Db.open(indexFile, Db.DB_BTREE, Db.DB_RDONLY, 0, this, null);
	docDb = Db.open(docFile, Db.DB_BTREE, Db.DB_RDONLY, 0, this, null);

	error = "";
	valid = true;
  }

  public boolean checkSyntax(Vector tokenlist)
  {
    tokenVector = tokenlist;
    valid = true;
	stack = new Stack();
    fullexpr(false);
    return valid;
  }

  public Vector parse(Vector tokenlist, Vector words)
  {
	searchWords = words;
    tokenVector = tokenlist;
	stack = new Stack();
    valid = true;
    fullexpr(true);

    ResultList result = (ResultList) stack.pop();
    if (result == null)  // Ouch!
	{
		valid = false;
		error = "Expected to have something to parse!";
		return  null;
	}
	return result;
  }

  public String getErrorMessage()
  {
	return error;
  }

  public boolean hadError()
  {
	return !valid;
  }
	
  void fullexpr(boolean output)
  {
	tokens = tokenVector.elements();
    lookahead = lexan();
    expr(output);
    if (valid && lookahead != DONE)
    {
		setError("end of expression");
    }
  }

  int	lexan()
  {
	if ( tokens.hasMoreElements() )
	{
		current = (WeightWord) tokens.nextElement();
		if ( current.word.equals("&") )
		{
			return '&';
		}
		if ( current.word.equals("|") )
		{
			return '|';
		}
		if ( current.word.equals("!") )
		{
			return '!';
		}
		if ( current.word.equals("(") )
		{
			return '(';
		}
		if ( current.word.equals(")") )
		{
			return ')';
		}
		return WORD;
	}
	else
	{
		return DONE;
	}
  }

  void expr(boolean output)
  {
    term(output);
    while (true)
    {
		if (match('|'))
		{
			term(output);
			if (output)
			{
				perform_or();
			}
		}
		else
		  break;
    }
    if (valid && lookahead == WORD)
    {
		setError("'AND' or 'OR'");
    }
  }

  void term(boolean output)
  {
    boolean isand;
    
    factor(output);
    while (true)
    {
		if ((isand = match('&')) || match('!'))
		{
			factor(output);
			if (output)
			{
				perform_and(isand);
			}
		}
		else
		  break;
    }
  }

  void factor(boolean output)
  {
    if (match('('))
    {
		expr(output);
		if (match(')'))
		{
			return;
		}
		else
		{
			setError("')'");
		}
    }
    else if (lookahead == WORD)
    {
		if (output)
		{
			perform_push();
		}
		lookahead = lexan();
    }
    else
    {
		setError("a search word");
    }
  }

  boolean match(int t)
  {
    if (lookahead == t)
    {
		lookahead = lexan();
		return true;
    }
    else
	  return false;
  }

  void setError(String expected)
  {
    if (valid)
    {
		valid = false;
		error = "Expected " + expected;
		if (lookahead == DONE || current == null)
		{
			error = error + " at the end";
		}
		else
		{
			error = error + " instead of '" + current.word + "'";
			switch (lookahead)
			{
			case '&':	error = error + " or 'AND'";	break;
			case '|':	error = error + " or 'OR'";	break;
			case '!':	error = error + " or 'NOT'";	break;
			}
		}
    }
  }

  void perform_push()
  {
	String	temp = current.word;
    String	data;
    ResultList	list = new ResultList();
    //WordRecord	wr;
    DocMatch	dm;

    stack.push(list);
    if (current.isIgnore)
    {
		//
		// This word needs to be ignored.  Make it so.
		//
		list.isIgnore = true;
		return;
    }
    temp = temp.toLowerCase(); // done in setupWords
    if (temp.length() > MAX_WORD_LENGTH)
	{
		temp = temp.substring (0, MAX_WORD_LENGTH);
	}
	byte[] b = temp.getBytes();
	Dbt wkey = new Dbt(b);
	wkey.set_size(b.length);
	Dbt wdata = new Dbt();
	wdata.set_flags(Db.DB_DBT_MALLOC);

	try
	{
		if (wordDb.get(null, wkey, wdata, 0) == 0)
		{
			DocMatch.addDbEntry(wdata.get_data(), list, searchWords);
		}
	}
	catch ( Exception ex )
	{
		error = ex.toString();
		valid = false;
	}
  }

  void perform_and(boolean isand)
  {
    ResultList		l1 = (ResultList) stack.pop();
    ResultList		l2 = (ResultList) stack.pop();
    ResultList		result = new ResultList();
    int			    i;
    DocMatch		dm, dm2, dm3;

    //
    // If either of the arguments is not present, we will use the other as
    // the result.
    //
    if (l1 == null && l2 != null)
    {
		stack.push(l2);
		return;
    }
    else if (l1 != null && l2 == null)
    {
		stack.push(l1);
		return;
    }
    else if (l1 == null && l2 == null)
    {
		stack.push(result);
		return;
    }
    
    //
    // If either of the arguments is set to be ignored, we will use the
    // other as the result.
    //
    if (l1.isIgnore)
    {
		stack.push(l2);
		return;
    }
    else if (l2.isIgnore)
    {
		stack.push(isand ? l1 : result);
		return;
    }
    
    stack.push(result);
    for (i = 0; i < l2.size(); i++)
    {
		dm = (DocMatch) l2.elementAt(i);
		dm2 = l1.find(dm.id);
 
		if (dm2 != null && isand)
		{
			//
			// Duplicate document.  We just need to add the scored together.
			//
			dm3 = new DocMatch(dm.id, dm.score + dm2.score, Math.min(dm.anchor, dm2.anchor), dm.searchWords);
			result.addElement(dm3);
		}
		if (dm2 == null && !isand)
		{
			result.addElement(dm);
		}
    }
  }

  void perform_or()
  {
    ResultList		l1 = (ResultList) stack.pop();
    ResultList		result = (ResultList) stack.lastElement();
    int			    i;
    DocMatch		dm, dm2;

    //
    // If either of the arguments is not present, we will use the other as
    // the results.
    //
    if (l1 == null && result != null)
    {
		return;
    }
    else if (l1 != null && result == null)
    {
		stack.push(l1);
		return;
    }
    else if (l1 == null && result == null)
    {
		stack.push(new ResultList());
		return;
    }
    
    //
    // If either of the arguments is set to be ignored, we will use the
    // other as the result.
    //
    if (l1.isIgnore)
    {
		return;
    }
    else if (result.isIgnore)
    {
		result = (ResultList) stack.pop();
		stack.push(l1);
		return;
    }
    
    for (i = 0; i < l1.size(); i++)
    {
		dm = (DocMatch) l1.elementAt(i);
		dm2 = result.find(dm.id);
		if (dm2 != null)
		{
			//
			// Duplicate document.  We just need to add the scores together
			//
			dm2.score += dm.score;
			if (dm.anchor < dm2.anchor)
			  dm2.anchor = dm.anchor;
		}
		else
		{
			result.addElement(dm);
		}
    }
  }

  public void fetchURL(DocMatch match)
  {
	try
	{
		Dbt ikey = new Dbt(match.key);
		ikey.set_size(match.key.length);
		Dbt idata = new Dbt();
		idata.set_flags(Db.DB_DBT_MALLOC);
		int err = indexDb.get(null, ikey, idata, 0);

		if ( err != 0 )
		{
			valid = false;
			error = "Databasecall returned " + err;
		} 
		match.setUrlKey(idata.get_data());
	}
	catch ( Exception ex )
	{
		error = ex.toString();
		valid = false;
	}
  }

  public void fetchDocumentRef(DocMatch match)
  {
	if ( match.getUrlKey() == null ) fetchURL(match);
	if ( match.getUrlKey() == null ) return;
	try
	{
		Dbt wkey = new Dbt(match.getUrlKey());
		wkey.set_size(match.getUrlKey().length);
		Dbt wdata = new Dbt();
		wdata.set_flags(Db.DB_DBT_MALLOC);
		int err = docDb.get(null, wkey, wdata, 0);

		if ( err != 0 )
		{
			valid = false;
			error = "Databasecall returned " + err;
		}
		match.setRef (new DocumentRef(wdata.get_data()));
	}
	catch ( Exception ex )
	{
		error = ex.toString();
		valid = false;
	}
  }
} // Parser

